@@ -2,27 +2,25 @@ r"""
 The main core of EasyTorch
 """
 
-import math as _math
 import os as _os
 from collections import OrderedDict as _ODict
-from typing import List as _List
 
 import torch as _torch
-from torch.utils.data import Dataset as _Dataset
 
+import easytorch.data as _etdata
 import easytorch.utils as _etutils
-from easytorch.config.state import *
+from easytorch.config.status import *
 from easytorch.metrics import metrics as _base_metrics
 from easytorch.utils.logger import *
 from easytorch.utils.tensorutils import initialize_weights as _init_weights
 from .vision import plotter as _log_utils
-import torch.distributed as _dist
+import math as _math
 
 _sep = _os.sep
 
 
 class ETTrainer:
-    def __init__(self, args=None, data_handle=None, **kw):
+    def __init__(self, args: dict):
         r"""
         args: receives the arguments passed by the ArgsParser.
         cache: Initialize all immediate things here. Like scores, loss, accuracies...
@@ -30,18 +28,12 @@ class ETTrainer:
         optimizer: Initialize our optimizers.
         """
         self.args = _etutils.FrozenDict(args)
-        self.data_handle = data_handle
-
         self.cache = _ODict()
         self.nn = _ODict()
-        self.device = _ODict({'gpu': args.get('gpu', 'cpu')})
+        self.device = _ODict()
         self.optimizer = _ODict()
 
-    def init_nn(self,
-                init_models=True,
-                init_weights=True,
-                init_optimizer=True,
-                set_device=True):
+    def init_nn(self, init_models=True, init_weights=True, init_optimizer=True, set_device=True):
         r"""
         Call to user implementation of:
             Initialize models.
@@ -56,8 +48,8 @@ class ETTrainer:
             for k, m in [(_k, _m) for _k, _m in self.nn.items() if isinstance(_m, _torch.nn.Module)]:
                 success(f'Total params in {k}:' f' {sum(p.numel() for p in m.parameters() if p.requires_grad)}')
 
-        if init_optimizer: self._init_optimizer()
         if init_weights: self._init_nn_weights()
+        if init_optimizer: self._init_optimizer()
         if set_device: self._set_device()
 
     def _init_nn_weights(self):
@@ -66,20 +58,13 @@ class ETTrainer:
         If path to pretrained weights are given, it will be used instead.
         """
         if self.args['pretrained_path'] is not None:
-            self.load_checkpoint(self.args['pretrained_path'],
-                                 self.args.get('load_model_state', True),
-                                 self.args.get('load_optimizer_state', True))
-
+            self.load_checkpoint(self.args['pretrained_path'])
         elif self.args['phase'] == 'train':
             _torch.manual_seed(self.args['seed'])
             for mk in self.nn:
                 _init_weights(self.nn[mk])
 
-    def load_checkpoint(self,
-                        full_path,
-                        load_model_state=True,
-                        load_optimizer_state=True,
-                        src=MYSELF):
+    def load_checkpoint(self, full_path, src=MYSELF):
         r"""
         Load checkpoint from the given path:
             If it is an easytorch checkpoint, try loading all the models.
@@ -91,19 +76,17 @@ class ETTrainer:
             chk = _torch.load(full_path, map_location='cpu')
 
         if chk.get('_its_origin_', 'Unknown').lower() == src:
-            if load_model_state:
-                for m in chk['models']:
-                    try:
-                        self.nn[m].module.load_state_dict(chk['models'][m])
-                    except:
-                        self.nn[m].load_state_dict(chk['models'][m])
+            for m in chk['models']:
+                try:
+                    self.nn[m].module.load_state_dict(chk['models'][m])
+                except:
+                    self.nn[m].load_state_dict(chk['models'][m])
 
-            if load_optimizer_state:
-                for m in chk['optimizers']:
-                    try:
-                        self.optimizer[m].module.load_state_dict(chk['optimizers'][m])
-                    except:
-                        self.optimizer[m].load_state_dict(chk['optimizers'][m])
+            for m in chk['optimizers']:
+                try:
+                    self.optimizer[m].module.load_state_dict(chk['optimizers'][m])
+                except:
+                    self.optimizer[m].load_state_dict(chk['optimizers'][m])
         else:
             mkey = list(self.nn.keys())[0]
             try:
@@ -123,19 +106,14 @@ class ETTrainer:
         Expects list of GPUS as [0, 1, 2, 3]., list of GPUS will make it use DataParallel.
         If no GPU is present, CPU is used.
         """
-        if self.args.get('use_ddp'):
-            for model_key in self.nn:
-                self.nn[model_key] = self.nn[model_key].to(self.device['gpu'])
-            for model_key in self.nn:
-                self.nn[model_key] = _torch.nn.parallel.DistributedDataParallel(self.nn[model_key],
-                                                                                device_ids=[self.device['gpu']])
-        elif len(self.args['gpus']) >= 1:
+        self.device['gpu'] = _torch.device("cpu")
+        if CUDA_AVAILABLE and len(self.args['gpus']) >= 1:
             self.device['gpu'] = _torch.device(f"cuda:{self.args['gpus'][0]}")
             if len(self.args['gpus']) >= 2:
                 for model_key in self.nn:
                     self.nn[model_key] = _torch.nn.DataParallel(self.nn[model_key], self.args['gpus'])
-            for model_key in self.nn:
-                self.nn[model_key] = self.nn[model_key].to(self.device['gpu'])
+        for model_key in self.nn:
+            self.nn[model_key] = self.nn[model_key].to(self.device['gpu'])
 
     def _init_optimizer(self):
         r"""
@@ -159,39 +137,48 @@ class ETTrainer:
         """
         return _base_metrics.ETAverages(num_averages=1)
 
-    def save_checkpoint(self,
-                        full_path,
-                        save_model_state=True,
-                        save_optimizer_state=True,
-                        src=MYSELF):
-
+    def save_checkpoint(self, full_path, src=MYSELF):
         checkpoint = {'_its_origin_': src}
-
-        if save_model_state:
-            for k in self.nn:
-                checkpoint['models'] = {}
-                try:
-                    checkpoint['models'][k] = self.nn[k].module.state_dict()
-                except:
-                    checkpoint['models'][k] = self.nn[k].state_dict()
-
-        if save_optimizer_state:
-            for k in self.optimizer:
-                checkpoint['optimizers'] = {}
-                try:
-                    checkpoint['optimizers'][k] = self.optimizer[k].module.state_dict()
-                except:
-                    checkpoint['optimizers'][k] = self.optimizer[k].state_dict()
+        for k in self.nn:
+            checkpoint['models'] = {}
+            try:
+                checkpoint['models'][k] = self.nn[k].module.state_dict()
+            except:
+                checkpoint['models'][k] = self.nn[k].state_dict()
+        for k in self.optimizer:
+            checkpoint['optimizers'] = {}
+            try:
+                checkpoint['optimizers'][k] = self.optimizer[k].module.state_dict()
+            except:
+                checkpoint['optimizers'][k] = self.optimizer[k].state_dict()
         _torch.save(checkpoint, full_path)
 
     def init_experiment_cache(self):
         r"""What scores you want to plot."""
+        r"""By default it plots Loss,Accuracy,F1,Precision,Recall."""
         self.cache['log_header'] = 'Loss,Accuracy'
 
         r"""This is for best model selection: """
         r"""It tells which metrics to monitor and either to maximize(F1 score), minimize(MSE)"""
         self.cache.update(monitor_metric='time', metric_direction='maximize')
 
+    def save_if_better(self, epoch, metrics):
+        r"""
+        Save the current model as best if it has better validation scores.
+        """
+        sc = getattr(metrics, self.cache['monitor_metric'])
+        if callable(sc):
+            sc = sc()
+
+        if (self.cache['metric_direction'] == 'maximize' and sc >= self.cache['best_score']) or (
+                self.cache['metric_direction'] == 'minimize' and sc <= self.cache['best_score']):
+            self.save_checkpoint(self.cache['log_dir'] + _sep + self.cache['checkpoint'])
+            self.cache['best_score'] = sc
+            self.cache['best_epoch'] = epoch
+            success(f"Best Model Saved!!! : {self.cache['best_score']}", self.args['verbose'])
+        else:
+            warn(f"Not best: {sc}, {self.cache['best_score']} in ep: {self.cache['best_epoch']}", self.args['verbose'])
+
     def iteration(self, batch):
         r"""
         Left for user to implement one mini-bath iteration:
@@ -225,22 +212,17 @@ class ETTrainer:
         """
         pass
 
-    def evaluation(self,
-                   epoch=1,
-                   mode='eval',
-                   dataset_list=None,
-                   save_pred=False) -> dict:
+    def evaluation(self, epoch=0, mode='eval', dataset_list=None, save_pred=False):
+        info(f'{mode} ...', self.args['verbose'])
 
         for k in self.nn:
             self.nn[k].eval()
 
         eval_avg, eval_metrics = self.new_averages(), self.new_metrics()
 
-        if dataset_list is None:
-            return {'averages': eval_avg, 'metrics': eval_metrics}
-
-        info(f'{mode} ...', self.args['verbose'])
-        loaders = [self.data_handle.get_loader(handle_key=mode, shuffle=False, dataset=d) for d in dataset_list]
+        _args = {**self.args}
+        _args['shuffle'] = False
+        loaders = [_etdata.ETDataLoader.new(mode=mode, dataset=d, **_args) for d in dataset_list if d is not None]
         with _torch.no_grad():
             for loader in loaders:
                 its = []
@@ -257,8 +239,7 @@ class ETTrainer:
                         its.append(it)
 
                     if self.args['verbose'] and len(dataset_list) <= 1 and lazy_debug(i, add=epoch):
-                        info(
-                            f" Itr:{i}/{len(loader)}, Averages:{it.get('averages').get()}, Metrics:{it.get('metrics').get()}")
+                        info(f" Itr:{i}/{len(loader)},{it.get('averages').get()},{it.get('metrics').get()}")
 
                 eval_metrics.accumulate(metrics)
                 eval_avg.accumulate(avg)
@@ -268,10 +249,9 @@ class ETTrainer:
                 if save_pred:
                     self.save_predictions(loader.dataset, self._reduce_iteration(its))
 
-        info(f"{self.cache['experiment_id']} {mode} Averages:{eval_avg.get()}, Metrics:{eval_metrics.get()}",
+        info(f" {self.cache['experiment_id']} {mode} metrics: {eval_avg.get()}, {eval_metrics.get()}",
              self.args['verbose'])
-
-        return {'averages': eval_avg, 'metrics': eval_metrics}
+        return eval_avg, eval_metrics
 
     def _reduce_iteration(self, its):
         reduced = {}.fromkeys(its[0].keys(), None)
@@ -302,125 +282,52 @@ class ETTrainer:
 
         return reduced
 
-    def _on_iteration_end(self, **kw):
+    def _on_epoch_end(self, epoch, **kw):
+        r"""
+        Any logic to run after an epoch ends.
+        """
+        pass
+
+    def _on_iteration_end(self, i, epoch, it):
         r"""
         Any logic to run after an iteration ends.
         """
         pass
 
-    def _check_validation_score(self, **kw):
-        r"""
-        Save the current model as best if it has better validation scores.
-        """
-        sc = kw['validation']['metrics'].extract(self.cache['monitor_metric'])
-        improved = False
-        if self.cache['metric_direction'] == 'maximize':
-            improved = sc > self.cache['best_val_score'] + self.args.get('score_delta', SCORE_DELTA)
-        elif self.cache['metric_direction'] == 'minimize':
-            improved = sc < self.cache['best_val_score'] - self.args.get('score_delta', SCORE_DELTA)
-        return {'improved': improved, 'score': sc}
-
-    def _stop_early(self, **kw):
+    def _stop_early(self, epoch, **kw):
         r"""
         Stop the training based on some criteria.
          For example: the implementation below will stop training if the validation
          scores does not improve within a 'patience' number of epochs.
         """
-        if self.args['patience'] and kw['epoch'] - self.cache['best_val_epoch'] >= self.args['patience']:
-            return True
-
-        if self.cache['metric_direction'] == 'maximize':
-            return self.cache['best_val_score'] == self.args.get('score_max', SCORE_MAX)
-        elif self.cache['metric_direction'] == 'minimize':
-            return self.cache['best_val_score'] == self.args.get('score_min', SCORE_MIN)
-
-        return False
+        return epoch - self.cache['best_epoch'] >= self.args.get('patience', 'epochs')
 
     def _save_progress(self, epoch):
         _log_utils.plot_progress(self.cache, experiment_id=self.cache['experiment_id'],
-                                     plot_keys=[LogKey.TRAIN_LOG, LogKey.VALIDATION_LOG],
-                                     epoch=epoch)
+                                 plot_keys=[LogKey.TRAIN_LOG, LogKey.VALIDATION_LOG], epoch=epoch)
 
-    def training_iteration(self, i, batch) -> dict:
+    def training_iteration(self, i, batch):
         r"""
         Learning step for one batch.
         We decoupled it so that user could implement any complex/multi/alternate training strategies.
         """
         it = self.iteration(batch)
         it['loss'].backward()
-        if i % self.args.get('grad_accum_iters', 1) == 0:
+        if i % self.args.get('num_iterations', 1) == 0:
             for optim in self.optimizer:
                 self.optimizer[optim].step()
                 self.optimizer[optim].zero_grad()
         return it
 
-    def reduce_scores(self, accumulator: list) -> dict:
-        averages = self.new_averages()
-        metrics = self.new_metrics()
-        if all([a is None for a in accumulator]):
-            return {f"averages": averages,
-                    f"metrics": metrics}
-
-        for acc in accumulator:
-            averages.accumulate(acc['averages'])
-            metrics.accumulate(acc['metrics'])
-
-        if self.args['use_ddp']:
-            avg_serial = _torch.tensor(averages.serialize()).to(self.device['gpu'])
-            _dist.reduce(avg_serial, dst=MASTER_RANK, op=_dist.ReduceOp.SUM)
-
-            metrics_serial = _torch.tensor(metrics.serialize()).to(self.device['gpu'])
-            _dist.reduce(metrics_serial, dst=MASTER_RANK, op=_dist.ReduceOp.SUM)
-
-            averages.reset()
-            averages.update(*avg_serial.cpu().numpy().tolist())
-
-            metrics.reset()
-            metrics.update(*metrics_serial.cpu().numpy().tolist())
-
-        return {f"averages": averages,
-                f"metrics": metrics}
-
-    def save_if_better(self, **kw):
-        val_check = self._check_validation_score(**kw)
-        if val_check['improved']:
-            self.save_checkpoint(self.cache['log_dir'] + _sep + self.cache['best_checkpoint'])
-            self.cache['best_val_score'] = val_check['score']
-            self.cache['best_val_epoch'] = kw['epoch']
-            success(f" *** Best Model Saved!!! *** : {self.cache['best_val_score']}", self.args['verbose'])
-        else:
-            info(
-                f"Not best: {val_check['score']}, {self.cache['best_val_score']} in ep: {self.cache['best_val_epoch']}",
-                self.args['verbose'])
-
-    def validation(self, epoch, val_dataset_list: _List[_Dataset]) -> dict:
-        return self.evaluation(epoch=epoch, mode='validation', dataset_list=val_dataset_list)
-
-    def _global_debug(self, running_averages, running_metrics, **kw):
-        """Update running accumulators."""
-        running_averages.accumulate(kw.get('averages'))
-        running_metrics.accumulate(kw.get('metrics'))
-
-        """Reset iteration accumulator"""
-        N = kw['num_iters']
-        i, e = kw['i'], kw['epoch']
-
-        if lazy_debug(i, add=e) or i == N:
-            info(
-                f"Ep:{e}/{self.args['epochs']},Itr:{i}/{N}, Averages:{running_averages.get()}, Metrics:{running_metrics.get()}",
-                self.args['verbose'])
-            r"""Debug and reset running accumulators"""
-
-            if not self.args['use_ddp']:
-                """Plot only in non-ddp mode to maintain consistency"""
-                self.cache[LogKey.TRAIN_LOG].append([*running_averages.get(), *running_metrics.get()])
-
-            running_averages.reset(), running_metrics.reset()
-
-    def train(self, train_dataset: _Dataset, val_dataset_list: _List[_Dataset]) -> None:
+    def train(self, dataset, val_dataset):
         info('Training ...', self.args['verbose'])
-        train_loader = self.data_handle.get_loader(handle_key='train', shuffle=True, dataset=train_dataset)
 
+        local_iter = self.args.get('num_iterations', 1)
+        _args = {**self.args}
+        _args['shuffle'] = True
+        train_loader = _etdata.ETDataLoader.new(mode='train', dataset=dataset, **_args)
+
+        tot_iter = len(train_loader) // local_iter
         for ep in range(1, self.args['epochs'] + 1):
             for k in self.nn:
                 self.nn[k].train()
@@ -429,62 +336,51 @@ class ETTrainer:
             its = []
 
             """Collect epoch metrics and averages"""
-            epoch_avg, epoch_metrics = self.new_averages(), self.new_metrics()
+            ep_avg, ep_metrics = self.new_averages(), self.new_metrics()
 
             """Keep track of running metrics and averages for logging/plotting"""
             _metrics, _avg = self.new_metrics(), self.new_averages()
-
-            if self.args.get('use_ddp'):
-                train_loader.sampler.set_epoch(ep)
-
-            num_iters = len(train_loader) // self.args['grad_accum_iters']
             for i, batch in enumerate(train_loader, 1):
                 its.append(self.training_iteration(i, batch))
+
                 """When end of iteration"""
-                if i % self.args['grad_accum_iters'] == 0:
+                if i % local_iter == 0:
                     it = self._reduce_iteration(its)
 
                     """Update global accumulators"""
-                    its = []
-                    it['num_iters'] = num_iters
-                    it['i'] = i // self.args['grad_accum_iters']
-                    epoch_avg.accumulate(it.get('averages'))
-                    epoch_metrics.accumulate(it.get('metrics'))
+                    ep_avg.accumulate(it.get('averages'))
+                    ep_metrics.accumulate(it.get('metrics'))
 
-                    if self.args['is_master']:
-                        self._global_debug(_avg, _metrics, epoch=ep, **it)
-                    self._on_iteration_end(i=i, epoch=ep, it=it)
+                    """Update running accumulators."""
+                    _avg.accumulate(it.get('averages'))
+                    _metrics.accumulate(it.get('metrics'))
+
+                    """Reset iteration accumulator"""
+                    _i, its = i // local_iter, []
+                    if lazy_debug(_i, add=ep) or _i == tot_iter:
+                        info(f"Ep:{ep}/{self.args['epochs']},Itr:{_i}/{tot_iter},{_avg.get()},{_metrics.get()}",
+                             self.args['verbose'])
+
+                        r"""Debug and reset running accumulators"""
+                        self.cache[LogKey.TRAIN_LOG].append([*_avg.get(), *_metrics.get()])
+                        _metrics.reset(), _avg.reset()
+
+                    self._on_iteration_end(i, ep, it)
 
             """Validation step"""
-            reduced_epoch = self.reduce_scores([{'averages': epoch_avg, 'metrics': epoch_metrics}])
-            epoch_out = {'epoch': ep, 'training': reduced_epoch}
-            if val_dataset_list:
-                val_out = self.validation(ep, val_dataset_list)
-                epoch_out['validation'] = self.reduce_scores([val_out])
+            val_averages, val_metric = self.evaluation(epoch=ep, mode='validation', dataset_list=[val_dataset])
+            self.cache[LogKey.VALIDATION_LOG].append([*val_averages.get(), *val_metric.get()])
+            self.save_if_better(ep, val_metric)
 
-            if self.args['is_master']:
-                self._global_epoch_end(**epoch_out)
+            self._on_epoch_end(epoch=ep, epoch_averages=ep_avg, epoch_metrics=ep_metrics,
+                               validation_averages=val_averages, validation_metric=val_metric)
 
-            self._on_epoch_end(**epoch_out)
-            if self._stop_early(**epoch_out):
+            if lazy_debug(ep, _math.log(ep)):
+                self._save_progress(epoch=ep)
+
+            if self._stop_early(epoch=ep, epoch_averages=ep_avg, epoch_metrics=ep_metrics,
+                                validation_averages=val_averages, validation_metric=val_metric):
                 break
 
         """Plot at the end regardless."""
         self._save_progress(epoch=ep)
-
-    def _global_epoch_end(self, **kw):
-        if kw.get('train') is not None:
-            self.cache[LogKey.TRAIN_LOG].append(
-                [*kw['train']['averages'].get(), *kw['train']['metrics'].get()]
-            )
-        if kw.get('validation') is not None:
-            self.save_if_better(**kw)
-            self.cache[LogKey.VALIDATION_LOG].append(
-                [*kw['validation']['averages'].get(), *kw['validation']['metrics'].get()]
-            )
-        if lazy_debug(kw['epoch'], _math.log(kw['epoch'])):
-            self._save_progress(epoch=kw['epoch'])
-
-    def _on_epoch_end(self, **kw):
-        """Local epoch end"""
-        pass
